import configparser
import glob
import inspect
import json
import os

import cv2
import numpy as np

from openmmla_vision.utils.input_utils import flush_input, get_function_calibrator
from openmmla_vision.utils.logger_utils import get_logger


class CameraCalibrator:
    logger = get_logger('camera-calibrator')

    def __init__(self, project_dir: str = None, config_path: str = None):
        # Set the project root directory
        if project_dir is None:
            caller_frame = inspect.stack()[1]
            caller_module = inspect.getmodule(caller_frame[0])
            if caller_module is None:
                self.project_dir = os.getcwd()
            else:
                self.project_dir = os.path.dirname(os.path.abspath(caller_module.__file__))
        else:
            self.project_dir = project_dir

        # Determine the configuration path
        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(self.project_dir, config_path)
        else:
            self.config_path = os.path.join(self.project_dir, 'conf/video_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        self.cameras_dir = os.path.join(self.project_dir, 'camera_calib/cameras')
        os.makedirs(self.cameras_dir, exist_ok=True)

        self.config = configparser.ConfigParser()
        self.config.read(self.config_path)
        self.res_width = int(self.config['DEFAULT']['res_width'])
        self.res_height = int(self.config['DEFAULT']['res_height'])

        # Checkerboard dimensions
        self.CHECKERBOARD = (6, 9)

        # Termination criteria
        self.subpix_criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)

        # Object points array
        self.objp = np.zeros((1, self.CHECKERBOARD[0] * self.CHECKERBOARD[1], 3), np.float32)
        self.objp[0, :, :2] = np.mgrid[0:self.CHECKERBOARD[0], 0:self.CHECKERBOARD[1]].T.reshape(-1, 2)

        # Lists to store object points and image points
        self.obj_points = []
        self.img_points = []

    def run(self):
        func_map = {1: self._capture_images, 2: self._calibrate_camera}
        while True:
            try:
                select_fun = get_function_calibrator()
                if select_fun == 0:
                    self.logger.info("Exiting video synchronizer...")
                    break
                func_map.get(select_fun, lambda: print("Invalid option."))()
            except (Exception, KeyboardInterrupt) as e:
                self.logger.warning("%s, Come back to the main menu.", e, exc_info=True)

    def _capture_images(self):
        resolution = [self.res_width, self.res_height]
        flush_input()
        camera_name = input("Enter the camera name for capturing images: ")
        saved_directory = os.path.join(self.cameras_dir, camera_name)
        os.makedirs(saved_directory, exist_ok=True)

        available_seeds = self._detect_video_seeds()
        camera_seed = self._choose_camera_seed(available_seeds)
        if camera_seed is None:
            self.logger.warning("No available camera seed found.")
            return

        cam = cv2.VideoCapture(camera_seed)
        cam.set(3, resolution[0])  # Set horizontal resolution
        cam.set(4, resolution[1])  # Set vertical resolution
        # cam.set(cv2.CAP_PROP_AUTOFOCUS, 1)
        # cam.set(cv2.CAP_PROP_SHARPNESS, 100) # 'value' ranges from 0 to the maximum sharpness value supported.

        try:
            self.capture_and_save_image(cam, saved_directory)
        finally:
            cam.release()
            cv2.destroyAllWindows()
            cv2.waitKey(1)

    def _detect_video_seeds(self):
        available_video_seeds = []
        number_of_detected_seeds = 0
        for i in range(4):
            cap = cv2.VideoCapture(i)
            if cap.isOpened():
                print(f"{number_of_detected_seeds} : Camera seed {i} is available.")
                number_of_detected_seeds += 1
                available_video_seeds.append(i)
            cap.release()

        if 'RTMP' in self.config and 'video_streams' in self.config['RTMP']:
            video_stream_list = self.config['RTMP']['video_streams'].split(',')
            for streams in video_stream_list:
                if streams:
                    print(f"{number_of_detected_seeds} : RTMP stream {streams} is available.")
                    available_video_seeds.append(streams)
                    number_of_detected_seeds += 1

        return available_video_seeds

    def _choose_camera_seed(self, available_video_seeds):
        if not available_video_seeds:
            return None
        while True:
            try:
                seed_id = int(input("Choose your video seed id: "))
                if 0 <= seed_id < len(available_video_seeds):
                    return available_video_seeds[seed_id]
                else:
                    self.logger.warning("Invalid selection. Please choose a valid video seed.")
            except ValueError:
                self.logger.warning("Please enter a valid number.")

    def _calibrate_camera(self):
        camera_name = self._select_camera()
        selected_path = os.path.join(self.cameras_dir, camera_name)
        is_fisheye = input("Is the camera a fisheye lens? (Y/n): ").lower() == 'y'

        try:
            images = glob.glob(os.path.join(selected_path, '*.jpg'))
            for img_file in images:
                img = cv2.imread(img_file)
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                ret, corners = cv2.findChessboardCorners(gray, self.CHECKERBOARD,
                                                         cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_FAST_CHECK + cv2.CALIB_CB_NORMALIZE_IMAGE)
                if ret:
                    self.obj_points.append(self.objp)
                    corners2 = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1),
                                                self.subpix_criteria) if not is_fisheye else corners
                    self.img_points.append(corners2)
                    cv2.drawChessboardCorners(img, self.CHECKERBOARD, corners2, ret)
                    cv2.imshow('img', img)
                    cv2.waitKey(0)
            cv2.destroyAllWindows()
            cv2.waitKey(1)

            if is_fisheye:
                ret, K, D, rvecs, tvecs = cv2.fisheye.calibrate(
                    self.obj_points, self.img_points, gray.shape[::-1], None, None,
                    flags=(cv2.fisheye.CALIB_RECOMPUTE_EXTRINSIC + cv2.fisheye.CALIB_CHECK_COND +
                           cv2.fisheye.CALIB_FIX_SKEW),
                    criteria=(cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 1e-6))
            else:
                ret, K, D, rvecs, tvecs = cv2.calibrateCamera(
                    self.obj_points, self.img_points, gray.shape[::-1], None, None)

            self._update_configuration(camera_name, K, D, is_fisheye)
        except Exception as e:
            self.logger.error("Error occurred during calibration: %s", e, exc_info=True)
        finally:
            self._clean_up()

    def _select_camera(self):
        directories = [d for d in os.listdir(self.cameras_dir) if os.path.isdir(os.path.join(self.cameras_dir, d))]
        for index, directory in enumerate(directories):
            print(f"{index}: {directory}")
        flush_input()
        choice = int(input("Enter the index of the camera folder you'd like to use: "))
        return directories[choice]

    def _clean_up(self):
        self.obj_points = []
        self.img_points = []

    def _update_configuration(self, camera_name, K, D, is_fisheye):
        k_list = K.tolist()
        d_list = D.tolist()
        k_str = json.dumps(k_list)
        d_str = json.dumps(d_list)
        params_str = json.dumps([K[0, 0], K[1, 1], K[0, 2], K[1, 2]])

        self.config[camera_name] = {
            'fisheye': str(is_fisheye),
            'params': params_str,
            'K': k_str,
            'D': d_str
        }

        with open(self.config_path, 'w') as configfile:
            self.config.write(configfile)
        print(f"Configuration updated for {camera_name}.")

    @staticmethod
    def capture_and_save_image(cam, directory):
        image_number = 1
        print("Press 'c' to capture the image, or 'q' to quit.")
        while cam.isOpened():
            result, image = cam.read()
            if result:
                cv2.imshow("Real-Time Capture", image)
                key = cv2.waitKey(1) & 0xFF

                if key == ord('c'):
                    filename = f"{directory}/{image_number}.jpg"
                    cv2.imwrite(filename, image)
                    print(f"Image captured as {filename}")
                    cv2.imshow("Captured Image", image)
                    cv2.waitKey(1)
                    flush_input()
                    user_input = input("Are you satisfied with the image? (Y/n): ")
                    if user_input.lower() == 'y':
                        image_number += 1
                        print(f"Image saved as {filename}")
                    else:
                        print("Image discarded. Continue capturing...")
                        os.remove(filename)
                    cv2.destroyWindow("Captured Image")
                elif key == ord('q'):
                    return
            else:
                print("No image detected. Please try again.")
                return



import re

from openmmla_vision.utils.image_utils import convert_image_to_base64


def generate_vlm_prompt_msg(id_position_dict=None, image_input=None):
    """Generate a prompt message for VLM.

    Args:
        id_position_dict: Dictionary containing the position of each person in the image
        image_input: Either a string (file path) or bytes (image data)

    Returns:
        Formatted prompt message
    """
    if not id_position_dict:
        user_text = (
            "Analyze this laboratory image:\n\n"
            "For each person you can identify in the image:\n"
            "1. Assign a unique ID (e.g., Person 1, Person 2, etc.)\n"
            "2. Describe only what you can see:\n"
            "   a) Approximate position in the image\n"
            "   b) Posture & orientation\n"
            "   c) Gaze direction\n"
            "   d) Hands & objects (if visible)\n"
            "   e) Clear interactions\n"
            "\nFormat:\n"
            "Person [X]:\n"
            "- Position: [brief location]\n"
            "- Visible details: [key observations]\n"
            "\nBe specific about what you can see and what you can't. "
            "If hands or any other part is not visible, state this explicitly. "
            "Do not make assumptions about unseen elements. "
            "Analyze each person you can identify in the image, one by one."
        )
    else:
        user_text = (
            f"Analyze this laboratory image with {len(id_position_dict)} individuals:\n"
        )

        for person_id, position in id_position_dict.items():
            user_text += f"- ID {person_id}: pos {position}\n"

        user_text += (
            "\nFor each person:\n"
            "1. Confirm ID based on position.\n"
            "2. Describe only what you can see:\n"
            "   a) Posture & orientation\n"
            "   b) Gaze direction\n"
            "   c) Hands & objects (if visible)\n"
            "   d) Clear interactions\n"
            "\nFormat:\n"
            "ID [X]:\n"
            "- Position: [brief location]\n"
            "- Visible details: [key observations]\n"
            "\nBe specific about what you can see and what you can't. "
            "If hands or any other part is not visible, state this explicitly. "
            "Do not make assumptions about unseen elements."
        )

    user_text += " Take a deep breath and do the analysis step by step."

    system_content = (
        "You're a capable video ethnographer analyzing a lab experiment."
    )

    if image_input:  # OpenAI API format message
        image_b64 = convert_image_to_base64(image_input)
        vlm_messages = [
            {"role": "system", "content": system_content},
            {"role": "user", "content": [{"type": "text", "text": user_text},
                                         {"type": "image_url", "image_url": {"url": image_b64}}]}
        ]
    else:
        vlm_messages = [
            {"role": "system", "content": system_content},
            {"role": "user", "content": f"(<image>./</image>)\n{user_text}"}
        ]

    return vlm_messages


def generate_llm_prompt_msg(image_description, action_definitions):
    """Generate a prompt message for LLM.

    Args:
        image_description: Description of the image generated by VLM
        action_definitions: Definitions of the actions to categorize

    Returns:
        Formatted prompt message
    """
    user_text = (f"You will now assist with analysing an image description text. Your task is to first categorize each"
                 f"individual's action from the given image description into one of the defined action classes from"
                 f"the action definitions. Then, format the categorization results into a target format.\n\n"
                 f"Image Description:\n{image_description}\n\n"
                 f"Action Definitions:\n{action_definitions}\n\n"
                 f"Carefully read the image description and action definitions and proceed with the following instructions:\n"
                 f"1. Categorize their action into one of the defined classes.\n"
                 f"2. Format your categorization results into a dictionary string with <ID> : <Action> key value pairs,"
                 f" for example, {{'1' : 'Action_1', '7' : 'Action_2'}}.\n"
                 f"Take a deep breath and do the analysis step by step.")

    system_content = (
        "You are now a taxonomist and sociologist who is good at categorizing human behaviors based on the given "
        "context of image description and actions definition. ")

    llm_messages = [
        {"role": "system", "content": system_content},
        {"role": "user", "content": user_text}
    ]

    return llm_messages


def parse_text_to_dict(ai_message: str):
    """Parse a string containing key-value pairs and return a dictionary.

    Args:
        ai_message: The input string to parse.
    Returns:
        A dictionary with keys as ids and values as action descriptions.
    """
    pattern = r"[\"'`]?(\d+)[\"'`]?\s*:\s*[\"'`]?([\w\s]+)[\"'`]?"
    matches = re.finditer(pattern, ai_message)
    return {match.group(1).strip(): match.group(2).strip() for match in matches}



import os

import cv2
import numpy as np

from openmmla_vision.utils.image_utils import load_image


def detect_apriltags(image_input, tag_detector, render=True, show=True, save=False, save_path=None):
    """
    Detect the apriltags in an image and return the positions of the tags.

    Args:
        image_input: Either a string (file path) or bytes (image data)
        tag_detector: AprilTag detector object
        render: Render the detected tags on the image or not
        show: Show the detected image or not
        save: Save the detected image or not
        save_path: Path to save the detected image (only used if save=True and image_input is bytes)
    """
    tag_pos = {}

    # Load image
    image = load_image(image_input)

    # Image resolution
    height, width, _ = image.shape
    print(f"Image resolution: {width}x{height} (Width x Height)")

    # Convert to grayscale
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    # Detect tags in the image
    tags = tag_detector.detect(gray)

    # Iterate over detected tags
    for tag in tags:
        # Calculate the center of the tag
        center = np.mean(tag.corners, axis=0)
        corners = np.int32(tag.corners)

        # Draw the tag ID with a background and border
        if render:
            tag_position = (int(corners[0][0]), int(corners[0][1]) - 10)  # Adjust position above the tag
            text = f"{tag.tag_id}"
            font = cv2.FONT_HERSHEY_SIMPLEX
            font_scale = 5
            font_thickness = 14
            text_size, _ = cv2.getTextSize(text, font, font_scale, font_thickness)
            text_x, text_y = tag_position
            # Draw black border
            cv2.rectangle(image, (text_x - 20, text_y - text_size[1] - 20), (text_x + text_size[0] + 10, text_y + 30),
                          (0, 0, 0), -1)
            # Draw white background
            cv2.rectangle(image, (text_x - 10, text_y - text_size[1] - 10), (text_x + text_size[0], text_y + 20),
                          (255, 255, 255), -1)
            # Put black text
            cv2.putText(image, text, (text_x, text_y), font, font_scale, (0, 0, 0), font_thickness)

        # Print the center position
        x = float(f'{center[0] / width:.4f}')
        y = float(f'{center[1] / height:.4f}')
        print(f"Person ID {tag.tag_id} center position: [{x}, {y}]")
        tag_pos[tag.tag_id] = [x, y]

    # Show the result
    if show:
        cv2.imshow('AprilTag Detection', image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    # Save detected image
    if save:
        if isinstance(image_input, str):
            dirname, filename = os.path.split(image_input)
            name, ext = os.path.splitext(filename)
            save_path = os.path.join(dirname, f"{name}_detected{ext}")
        elif save_path is None:
            raise ValueError("save_path must be provided when saving image data")

        cv2.imwrite(save_path, image)
        print(f"Detected image saved as {save_path}")

    return tag_pos



import base64

import cv2
import numpy as np


def load_image(image_input):
    """
    Load an image from either a file path or image data.

    Args:
        image_input: Either a string (file path) or bytes (image data)

    Returns:
        numpy.ndarray: The loaded image
    """
    if isinstance(image_input, str):
        # It's a file path
        image = cv2.imread(image_input)
    elif isinstance(image_input, bytes):
        # It's image data
        nparr = np.frombuffer(image_input, np.uint8)
        image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    else:
        raise ValueError("Invalid input type. Expected string (file path) or bytes (image data).")

    if image is None:
        raise ValueError("Failed to load image")

    return image


def convert_image_to_base64(image_input):
    """Convert an image to base64 format.

    Args:
        image_input: Either a string (file path) or bytes (image data)

    Returns:
        str: The image in base64 format
    """
    if isinstance(image_input, str):
        with open(image_input, 'rb') as file:
            image_data = file.read()
    elif isinstance(image_input, bytes):
        image_data = image_input
    else:
        raise ValueError("Invalid input type. Expected string (file path) or bytes (image data).")

    return "data:image/jpeg;base64," + base64.b64encode(image_data).decode('utf-8')



import configparser
import inspect
import json
import os
import time

import cv2
import numpy as np
from pupil_apriltags import Detector

from openmmla_vision.utils.input_utils import get_function_base
from openmmla_vision.utils.logger_utils import get_logger
from openmmla_vision.utils.mqtt_client import MQTTClientWrapper
from openmmla_vision.utils.stream_utils import WebcamVideoStream


class CameraTagDetector:
    """Camera detector class for detecting AprilTags from camera feed"""
    logger = get_logger('camera-tag-detector')

    def __init__(self, project_dir: str = None, config_path: str = None, max_badge_id=15):
        # Set the project root directory
        if project_dir is None:
            caller_frame = inspect.stack()[1]
            caller_module = inspect.getmodule(caller_frame[0])
            if caller_module is None:
                self.project_dir = os.getcwd()
            else:
                self.project_dir = os.path.dirname(os.path.abspath(caller_module.__file__))
        else:
            self.project_dir = project_dir

        # Determine the configuration path
        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(self.project_dir, config_path)
        else:
            self.config_path = os.path.join(self.project_dir, 'conf/video_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        self.max_badge_id = max_badge_id

        self.cameras_dir = os.path.join(self.project_dir, 'camera_calib/cameras')

        self.camera_info = {}
        self.camera_seed = None
        self.sender_id = None
        self.video_stream = None
        self.camera_configured = False

        self.config = configparser.ConfigParser()
        self.config.read(self.config_path)
        self.tag_size = float(self.config['DEFAULT']['tag_size'])
        self.families = self.config['DEFAULT']['families']
        self.res_width = int(self.config['DEFAULT']['res_width'])
        self.res_height = int(self.config['DEFAULT']['res_height'])
        self.res = (self.res_width, self.res_height)

        self.detector = Detector(families=self.families, nthreads=4)
        self.mqtt_client = MQTTClientWrapper(self.config_path)

    def run(self):
        print('\033]0;Camera Detector\007')
        func_map = {1: self._start, 2: self._set_camera}

        while True:
            try:
                select_fun = get_function_base()
                if select_fun == 0:
                    self.logger.info("Exiting video base...")
                    break
                func_map.get(select_fun, lambda: print("Invalid option."))()
            except (Exception, KeyboardInterrupt) as e:
                self.logger.warning("%s, Come back to the main menu.", e, exc_info=True)

    def _start(self):
        """Start AprilTag detection"""
        if not self.camera_configured:
            self.logger.warning("Camera is not configured.")
            return self._set_camera()

        # MQTT client reinitialization
        self.mqtt_client.reinitialise()
        self.mqtt_client.loop_start()

        # Start video stream
        self._configure_video_capture(self.camera_seed)

        try:
            self._process_frames()
        except (Exception, KeyboardInterrupt) as e:
            self.logger.warning("%s, capture interrupted.", e, exc_info=False)
        finally:
            self.video_stream.stop()
            self.mqtt_client.loop_stop()
            cv2.destroyAllWindows()
            cv2.waitKey(1)

    def _set_camera(self):
        self.camera_configured = False

        self.camera_info = self._configure_camera_params()
        if self.camera_info is None:
            self.logger.warning("Camera configuration failed, please calibrate your camera first.")
            return

        available_seeds = self._detect_video_seeds()
        self.camera_seed = self.choose_camera_seed(available_seeds)
        if self.camera_seed is None:
            self.logger.warning("No available camera seed found.")
            return

        self.sender_id = input("Input your sender id, 'm' for main camera, and 'a', 'b', 'c', 'd' for alternatives "
                               "camera: ")
        self.camera_configured = True
        print(f'\033]0;Camera Detector {self.sender_id}\007')

    def _configure_camera_params(self):
        camera_choices = [d for d in os.listdir(self.cameras_dir) if os.path.isdir(os.path.join(self.cameras_dir, d))]
        for idx, choice in enumerate(camera_choices):
            print(f"{idx}: {choice}")

        if not camera_choices:
            return None

        while True:
            try:
                selection = int(input("Choose your camera name with number: "))
                if not 0 <= selection < len(camera_choices):
                    self.logger.warning("Invalid selection. Please choose a valid number.")
                else:
                    chosen_camera = camera_choices[selection]
                    break
            except ValueError:
                self.logger.warning("Please enter a valid number.")

        fisheye = self.config.getboolean(chosen_camera, 'fisheye')
        params = json.loads(self.config.get(chosen_camera, 'params'))
        camera_info = {"fisheye": fisheye, "params": params, "res": self.res}

        if fisheye:
            K = np.array(json.loads(self.config.get(chosen_camera, 'K')))
            D = np.array(json.loads(self.config.get(chosen_camera, 'D')))
            map_1, map_2 = cv2.fisheye.initUndistortRectifyMap(K, D, np.eye(3), K, self.res, cv2.CV_16SC2)
            camera_info.update({"K": K, "D": D, "map_1": map_1, "map_2": map_2})

        return camera_info

    def _configure_video_capture(self, camera_seed):
        self.video_stream = WebcamVideoStream(format='MJPG', src=camera_seed, res=(1920, 1080))
        self.video_stream.start()

    def _process_frames(self):
        while True:
            frame = self.video_stream.read()
            if self.camera_info.get("fisheye", False):
                frame = cv2.remap(frame, self.camera_info["map_1"], self.camera_info["map_2"],
                                  interpolation=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT)

            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            results = self.detector.detect(gray, estimate_tag_pose=True, camera_params=self.camera_info["params"],
                                           tag_size=self.tag_size)

            tags = {}

            for tag in results:
                if int(tag.tag_id) > self.max_badge_id:
                    continue

                corners = np.int32(tag.corners)
                normal, tag = self.get_outward_normal(tag)
                tags[tag.tag_id] = [list(tag.pose_R.tolist()), list(tag.pose_t.tolist())]

                # Drawing annotations
                tag_center = np.mean(corners, axis=0)
                arrow_dir = normal[:2]
                scale_factor = 50
                end_point = tag_center + scale_factor * arrow_dir
                cv2.arrowedLine(frame, tuple(np.int32(tag_center)), tuple(np.int32(end_point)), (0, 0, 255), 2)
                cv2.polylines(frame, [corners], True, (0, 255, 0), thickness=2)
                cv2.putText(frame, str(tag.tag_id), org=(int(tag_center[0]) + 10, int(tag_center[1]) + 10),
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.8, color=(0, 255, 0), thickness=2)
                cv2.putText(frame, f"Rot: {tag.pose_R}",
                            (tag.corners[0][0].astype(int), tag.corners[0][1].astype(int) - 30),
                            cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255, 0, 0), 2)
                cv2.putText(frame, f"Trans: {tag.pose_t}",
                            (tag.corners[0][0].astype(int), tag.corners[0][1].astype(int) - 60),
                            cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255, 0, 0), 2)

            display_frame = cv2.resize(frame, (960, 540))
            cv2.imshow(f'AprilTags Detection from camera {self.sender_id}', display_frame)

            message = {
                "sender_id": self.sender_id,
                "tags": tags,
                "timestamp": time.time(),
            }
            message_str = json.dumps(message)
            self.mqtt_client.publish("camera/synchronize", message_str, qos=0, retain=False)

            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

    @staticmethod
    def get_outward_normal(tag):
        """Get the normal vector of the tag which perpendicular to the surface"""
        normal = np.dot(tag.pose_R, np.array([0, 0, 1])).ravel()
        if np.dot(normal, tag.pose_t.ravel()) < 0:
            tag.pose_R[:, 2] = -tag.pose_R[:, 2]  # Flip the sign of the third column of the rotation matrix
        else:
            normal = -normal
        return normal, tag

    def _detect_video_seeds(self):
        available_video_seeds = []
        number_of_detected_seeds = 0
        for i in range(4):
            cap = cv2.VideoCapture(i)
            if cap.isOpened():
                print(f"{number_of_detected_seeds} : Camera seed {i} is available.")
                number_of_detected_seeds += 1
                available_video_seeds.append(i)
            cap.release()

        if 'RTMP' in self.config and 'video_streams' in self.config['RTMP']:
            video_stream_list = self.config['RTMP']['video_streams'].split(',')
            for streams in video_stream_list:
                if streams:
                    print(f"{number_of_detected_seeds} : RTMP stream {streams} is available.")
                    available_video_seeds.append(streams)
                    number_of_detected_seeds += 1

        return available_video_seeds

    @staticmethod
    def choose_camera_seed(available_video_seeds):
        if not available_video_seeds:
            return None
        while True:
            try:
                seed_id = int(input("Choose your video seed id: "))
                if 0 <= seed_id < len(available_video_seeds):
                    print(f"Selected video seed: {available_video_seeds[seed_id]}")
                    return available_video_seeds[seed_id]
                else:
                    print("Invalid selection. Please choose a valid video seed.")
            except ValueError:
                print("Please enter a valid number.")



import configparser
import os

from flask import request, jsonify
from openai import OpenAI
from pupil_apriltags import Detector

from openmmla_vision.utils.apriltag_utils import detect_apriltags
from openmmla_vision.utils.logger_utils import get_logger
from openmmla_vision.utils.prompt_utils import parse_text_to_dict, generate_llm_prompt_msg, generate_vlm_prompt_msg


class VideoFrameAnalyzer:
    """Video frame analyzer receives images, processes them with VLM and LLM, and returns the results."""

    def __init__(self, project_dir, config_path=None):
        """Initialize the video frame analyzer.
        Args:
            project_dir: the project directory.
        """
        self.server_logger_dir = os.path.join(project_dir, 'server/logger')
        self.server_file_folder = os.path.join(project_dir, 'server/temp')
        os.makedirs(self.server_logger_dir, exist_ok=True)
        os.makedirs(self.server_file_folder, exist_ok=True)

        self.logger = get_logger('VideoFrameAnalyzer',
                                 os.path.join(self.server_logger_dir, 'video_frame_analyzer_server.log'))

        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(project_dir, config_path)
        else:
            self.config_path = os.path.join(project_dir, 'conf/video_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        config = configparser.ConfigParser()
        config.read(self.config_path)

        self.families = config['DEFAULT']['families']
        self.backend = config['Server']['backend']
        self.vlm_model = config['Server']['vlm_model']
        self.llm_model = config['Server']['llm_model']
        self.top_p = float(config['Server']['top_p'])
        self.temperature = float(config['Server']['temperature'])

        if self.backend == 'ollama':
            self.api_key = config['ollama']['api_key']
            self.vlm_base_url = config['ollama']['base_url']
            self.llm_base_url = config['ollama']['base_url']
        elif self.backend == 'vllm':
            self.api_key = config['vllm']['api_key']
            self.vlm_base_url = config['vllm']['vlm_base_url']
            self.llm_base_url = config['vllm']['llm_base_url']
        else:
            raise ValueError(f"Unsupported backend: {self.backend}")

        self.detector = Detector(families=self.families, nthreads=4)

        self.vlm_client = OpenAI(
            api_key=self.api_key,
            base_url=self.vlm_base_url,
        )

        self.llm_client = OpenAI(
            api_key=self.api_key,
            base_url=self.llm_base_url,
        )

        # Load action definitions from config
        self.action_definitions = '\n'.join([f"'{key}': {value}" for key, value in config['ActionDefinitions'].items()])

        # Load VLM extra body from config
        self.vlm_extra_body = {}
        if 'VLMExtraBody' in config:
            for key, value in config['VLMExtraBody'].items():
                if key == 'stop_token_ids':
                    self.vlm_extra_body[key] = [int(token) for token in value.split(',')]
                else:
                    self.vlm_extra_body[key] = value

    def process_image(self):
        """Process the image.

        Returns:
            A tuple containing the response and status code.
        """
        if 'image' not in request.files:
            return jsonify({'error': 'No image file provided'}), 400

        try:
            image_file = request.files['image']
            image_data = image_file.read()

            # Detect AprilTags
            id_positions = detect_apriltags(image_data, self.detector, show=False)

            # Process with VLM
            vlm_messages = generate_vlm_prompt_msg(id_positions, image_data)
            vlm_response = self.vlm_client.chat.completions.create(
                model=self.vlm_model,
                messages=vlm_messages,
                top_p=self.top_p,
                temperature=self.temperature,
                extra_body=self.vlm_extra_body
            )
            image_description = vlm_response.choices[0].message.content

            # Process with LLM
            llm_messages = generate_llm_prompt_msg(image_description, self.action_definitions)
            llm_response = self.llm_client.chat.completions.create(
                model=self.llm_model,
                messages=llm_messages,
                top_p=self.top_p,
                temperature=self.temperature
            )
            categorization_result = llm_response.choices[0].message.content
            parsed_result = parse_text_to_dict(categorization_result)

            return jsonify({
                'image_description': image_description,
                'categorization_result': parsed_result
            }), 200

        except Exception as e:
            self.logger.error(f"Error processing image: {str(e)}")
            return jsonify({"error": str(e)}), 500



import gc
import inspect
import json
import os
import queue
import time
from itertools import cycle

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation

from openmmla_vision.utils.input_utils import get_function_sync_manager
from openmmla_vision.utils.logger_utils import get_logger
from openmmla_vision.utils.mqtt_client import MQTTClientWrapper
from openmmla_vision.utils.transform_utils import (direct_transform_matrices, average_transform_matrices,
                                                   transform_point,
                                                   distance_between_points, transform_rotation,
                                                   distance_between_rotations,
                                                   export_main_transformations_json)

matplotlib.use('TkAgg')


class CameraSyncManager:
    """Class for synchronize the coordinate system between main camera and alternative camera, calculate out the
    transform matrix for each main/alter pairs."""
    logger = get_logger('camera-sync-manager')
    colors = cycle('bgrcmyk')

    def __init__(self, project_dir: str = None, config_path: str = None, sync: bool = True, top_K: int = 5000,
                 distance_threshold: float = 0.1, angle_threshold: float = 3.0, time_threshold_sync: float = 0.02,
                 time_threshold_unsync: float = 0.1):
        # Set the project root directory
        if project_dir is None:
            caller_frame = inspect.stack()[1]
            caller_module = inspect.getmodule(caller_frame[0])
            if caller_module is None:
                self.project_dir = os.getcwd()
            else:
                self.project_dir = os.path.dirname(os.path.abspath(caller_module.__file__))
        else:
            self.project_dir = project_dir

        # Determine the configuration path
        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(self.project_dir, config_path)
        else:
            self.config_path = os.path.join(self.project_dir, 'conf/video_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        self.sync = sync
        self.top_K = top_K
        self.distance_threshold = distance_threshold
        self.angle_threshold = angle_threshold
        self.time_threshold_sync = time_threshold_sync
        self.time_threshold_unsync = time_threshold_unsync
        self.time_threshold = self.time_threshold_sync if self.sync else self.time_threshold_unsync

        self.camera_sync_dir = os.path.join(self.project_dir, 'camera_sync')
        os.makedirs(self.camera_sync_dir, exist_ok=True)

        self.main_camera_id, self.alternative_camera_id = '', ''
        self.main_positions, self.main_rotations, self.main_last_update = {}, {}, {}
        self.alternative_positions, self.alternative_rotations, self.alternative_last_update = {}, {}, {}
        self.rotation_matrices, self.translations_matrices = [], []
        self.retrieved_R, self.retrieved_T = None, None
        self.fig, self.ax, self.anim = None, None, None
        self.plot_points_queue = queue.Queue()
        self.plot_points_list = []
        self.tag_color_map = {}

        self.mqtt_client = MQTTClientWrapper(self.config_path)

    def run(self):
        func_map = {1: self._start, 2: self._switch, 3: self._set_camera_id, 4: self._export_transformations,
                    5: self._clear_transformations}
        while True:
            try:
                select_fun = get_function_sync_manager()
                if select_fun == 0:
                    self.logger.info("Exiting video synchronizer...")
                    break
                func_map.get(select_fun, lambda: print("Invalid option."))()
            except (Exception, KeyboardInterrupt) as e:
                self.logger.warning("%s, Come back to the main menu.", e, exc_info=True)

    def _start(self):
        try:
            if not self.main_camera_id or not self.alternative_camera_id:
                return self._set_camera_id()

            self.mqtt_client.reinitialise(on_message=self._on_message, topics="camera/synchronize")
            self.mqtt_client.loop_start()
            self.fig, self.ax = plt.subplots(1, 1, subplot_kw={'projection': '3d'})
            plt.get_current_fig_manager().set_window_title(
                f"Camera Sync Manager: main:{self.main_camera_id} - alternative:{self.alternative_camera_id}")
            self.anim = FuncAnimation(self.fig, self._animate_plot, interval=10, cache_frame_data=False)
            plt.show()
        except KeyboardInterrupt:
            self.logger.info("Ctrl+C, exiting video synchronizer...")
        finally:
            self.mqtt_client.loop_stop()
            self._clean_up()

    def _switch(self):
        self.sync = not self.sync
        self.time_threshold = self.time_threshold_sync if self.sync else self.time_threshold_unsync
        print(f"Switching the synchronization mode to {self.sync} with time threshold {self.time_threshold}.")

    def _set_camera_id(self):
        self.main_camera_id = input("Please enter the main camera ID: ")
        self.alternative_camera_id = input("Please enter the alternative camera ID: ")
        print(f"\033]0;Camera Sync Manager: main:{self.main_camera_id} - alternative:{self.alternative_camera_id}\007")

    def _clean_up(self):
        """Reinitialize the variables and plot based on sync mode."""
        self.main_positions, self.main_rotations, self.main_last_update = {}, {}, {}
        self.alternative_positions, self.alternative_rotations, self.alternative_last_update = {}, {}, {}
        self.rotation_matrices, self.translations_matrices = [], []
        self.retrieved_R, self.retrieved_T = None, None
        self.plot_points_queue = queue.Queue()
        self.plot_points_list = []
        self.tag_color_map = {}

        if self.fig:
            plt.close(self.fig)

        gc.collect()

    def _animate_plot(self, i):
        """Update the plot based on points data queue."""
        self.ax.clear()
        try:
            plot_points_list = self.plot_points_queue.get_nowait()
            for main_point, transformed_point, color, tag in plot_points_list:
                self.ax.scatter(main_point[:, 0], main_point[:, 1], main_point[:, 2], c=color, marker='o',
                                label=f'Main Camera (Tag {tag})')
                self.ax.scatter(transformed_point[:, 0], transformed_point[:, 1], transformed_point[:, 2], c=color,
                                marker='x',
                                label=f'Alternative Camera (Transformed, Tag {tag})')
            self.ax.set_xlabel('X')
            self.ax.set_ylabel('Y')
            self.ax.set_zlabel('Z')
            self.ax.legend()
        except queue.Empty:
            pass

    def _on_message(self, client, userdata, msg):
        message = json.loads(msg.payload)

        sender_id = message["sender_id"]
        tags = message["tags"]
        current_time = time.time()

        for tag in list(self.main_positions.keys()):
            if (current_time - self.main_last_update.get(tag, 0)) > self.time_threshold:
                del self.main_positions[tag]
                del self.main_last_update[tag]

        for tag in list(self.alternative_positions.keys()):
            if (current_time - self.alternative_last_update.get(tag, 0)) > self.time_threshold:
                del self.alternative_positions[tag]
                del self.alternative_last_update[tag]

        for tag_id, tag_data in tags.items():
            position = tag_data[1]
            rotation = tag_data[0]

            if sender_id == self.main_camera_id:
                self.main_positions[tag_id] = position
                self.main_rotations[tag_id] = rotation
                self.main_last_update[tag_id] = current_time
            elif sender_id == self.alternative_camera_id:
                self.alternative_positions[tag_id] = position
                self.alternative_rotations[tag_id] = rotation
                self.alternative_last_update[tag_id] = current_time

            if tag_id in self.main_positions and tag_id in self.alternative_positions:
                color = self.tag_color_map.setdefault(tag_id, next(self.colors))
                if self.sync:
                    self._synchronize(tag_id, color)
                else:
                    self._validate(tag_id, color)

        if self.plot_points_list:
            self.plot_points_queue.put(self.plot_points_list)
            self.plot_points_list = []

    def _synchronize(self, tag, color):
        R, T = direct_transform_matrices(self.main_rotations[tag], self.main_positions[tag],
                                         self.alternative_rotations[tag], self.alternative_positions[tag])
        self.rotation_matrices.append(R)
        self.translations_matrices.append(T)
        self.rotation_matrices = self.rotation_matrices[-self.top_K:]
        self.translations_matrices = self.translations_matrices[-self.top_K:]

        R_avg, T_avg = average_transform_matrices(self.rotation_matrices, self.translations_matrices)
        print(f"R matrix for tag {tag}: {R_avg}, T vector for tag {tag}: {T_avg}")

        transformed_point = transform_point(self.alternative_positions[tag], R_avg, T_avg)
        transformed_rotation = transform_rotation(R_avg, self.alternative_rotations[tag])
        point_distance = distance_between_points(transformed_point, self.main_positions[tag])
        rotation_distance = distance_between_rotations(transformed_rotation, self.main_rotations[tag])

        print(
            f"Main position for tag {tag}: {self.main_positions[tag]}, "
            f"alternative position for tag {tag}: {self.alternative_positions[tag]}, "
            f"converted position for tag {tag}: {transformed_point}")
        print(f"Distances for tag {tag}: {point_distance} m, angles: {rotation_distance} °")

        if point_distance > self.distance_threshold or rotation_distance > self.angle_threshold:
            print(f"Outlier detected for tag {tag}. Rolling back to last valid transform matrices.")
            self.rotation_matrices.pop(-1)
            self.translations_matrices.pop(-1)
        else:
            self._update_json_file(R_avg, T_avg)

        self._add_points_to_list(tag, color, transformed_point)

    def _validate(self, tag, color):
        if self.retrieved_R is not None and self.retrieved_T is not None:
            transformed_point = transform_point(self.alternative_positions[tag], self.retrieved_R, self.retrieved_T)
            transformed_rotation = transform_rotation(self.retrieved_R, self.alternative_rotations[tag])
            point_distance = distance_between_points(transformed_point, self.main_positions[tag])
            rotation_distance = distance_between_rotations(transformed_rotation, self.main_rotations[tag])

            print(
                f"Main position for tag {tag}: {self.main_positions[tag]}, "
                f"alternative position for tag {tag}: {self.alternative_positions[tag]},"
                f"converted position for tag {tag}: {transformed_point}")
            print(f"Distances for tag {tag}: {point_distance} m, angles: {rotation_distance} °")

            self._add_points_to_list(tag, color, transformed_point)
        else:
            with open(os.path.join(self.camera_sync_dir, 'transformation_matrices.json'), 'r') as file:
                data = json.load(file)
            key = f'{self.alternative_camera_id}-{self.main_camera_id}'
            self.retrieved_R, self.retrieved_T = np.array(data[key]['R']), np.array(
                data[key]['T'])

    def _add_points_to_list(self, tag, color, transformed_point):
        main_point = np.array(self.main_positions[tag]).reshape(1, 3)
        transformed_point = np.array(transformed_point).reshape(1, 3)
        self.plot_points_list.append((main_point, transformed_point, color, tag))

    def _update_json_file(self, R, T):
        json_file_path = os.path.join(self.camera_sync_dir, 'transformation_matrices.json')
        data = None
        if os.path.exists(json_file_path):
            with open(json_file_path, 'r') as file:
                data = json.load(file)
        if data is None:
            data = {}

        key = f'{self.alternative_camera_id}-{self.main_camera_id}'
        data[key] = {"R": R, "T": T}
        with open(json_file_path, 'w') as file:
            json.dump(data, file, indent=4)

    def _export_transformations(self):
        main_system = input("Please enter the main system ID: ")
        input_path = os.path.join(self.camera_sync_dir, 'transformation_matrices.json')
        output_path = os.path.join(self.camera_sync_dir, f'transformation_matrices_{main_system}.json')
        export_main_transformations_json(input_path, output_path, main_system)

    def _clear_transformations(self):
        for file in os.listdir(self.camera_sync_dir):
            if file.startswith("transformation_matrices"):
                os.remove(os.path.join(self.camera_sync_dir, file))
                print(f"File {file} has been removed.")



from flask import Flask


def create_app(class_type, endpoint, method_name, class_args=None, route_args=None):
    """
    Create a Flask app for audio processing.

    Args:
        class_type (class): The class to instantiate (e.g., AudioEnhancer or AudioInferer).
        endpoint (str): The endpoint name for the Flask route.
        method_name (str): The method name to call on the instantiated class.
        class_args (dict): Additional arguments required to instantiate the class.
        route_args (dict): Arguments needed to define the route function.

    Returns:
        app (Flask): The Flask application.
    """
    app = Flask(__name__)

    if class_args is None:
        class_args = {}
    if route_args is None:
        route_args = {}

    processor = class_type(**class_args)

    @app.route(f'/{endpoint}', methods=['POST'])
    def process_route():
        method = getattr(processor, method_name)
        return method(**route_args)

    return app